from .metric import Metric


class HallucinationScore(Metric):
    """
    Calculates the hallucination score between two sets of answers.

    HallucinationScore computes the proportion of summaries where a question generated
    from the summary receives a 'Yes/No' answer from the summary, but an 'Unknown' answer
    from the source document. A high score indicates potential content in the summary
    that is absent from the source document.
    """

    @staticmethod
    def _compute_metric(answers_src, answers_sum, questions):
        """
        Computes the number of hallucinations between the answers from source and summary.

        Args:
            answers_src (dict): Answers derived from the source.
            answers_sum (dict): Answers derived from the summary.

        Returns:
            int: Number of questions indicating hallucinations.
        """
        answers_src_ls = list(answers_src.values())
        answers_sum_ls = list(answers_sum.values())

        halu_questions = []
        n_hallucination = 0

        for idx, (ans_src, ans_sum) in enumerate(zip(answers_src_ls, answers_sum_ls)):
            if ans_src.strip().lower() == "unknown" and ans_sum.strip().lower() in [
                "yes",
                "no",
            ]:
                n_hallucination += 1
                halu_question = questions[idx]
                halu_questions.append(f"{halu_question}")

        return n_hallucination, halu_questions

    @staticmethod
    def compute(answers_src, answers_sum, questions, n_questions):
        """
        Computes the hallucination score.

        Args:
            answers_src (dict): Answers derived from the source.
            answers_sum (dict): Answers derived from the summary.
            questions (dict): Questions generated from the summary.
            n_questions (int): Total number of questions.

        Returns:
            float: Hallucination score.
        """
        n_hallucination, halu_questions = HallucinationScore._compute_metric(
            answers_src, answers_sum, questions
        )
        halu_score = n_hallucination / n_questions
        explanation = halu_questions
        return halu_score, explanation



import pkg_resources
import requests
from dataclasses import asdict
from retrying import retry
from typing import List, Optional, Dict
from athina.errors.exceptions import NoAthinaApiKeyException
from athina.interfaces.athina import (
    AthinaFilters,
    AthinaInference,
    AthinaEvalRequestCreateRequest,
    AthinaEvalResultCreateRequest,
    AthinaExperiment,
)
from athina.interfaces.result import EvalPerformanceReport
from athina.keys import AthinaApiKey
from athina.helpers.constants import API_BASE_URL
from athina.errors.exceptions import CustomException

SDK_VERSION = pkg_resources.get_distribution("athina").version


class AthinaApiService:
    @staticmethod
    def _headers():
        athina_api_key = AthinaApiKey.get_key()
        return {
            "athina-api-key": athina_api_key,
        }

    @staticmethod
    def fetch_inferences(
        filters: Optional[AthinaFilters], limit: int
    ) -> List[AthinaInference]:
        """
        Load data from Athina API.
        """
        try:
            endpoint = f"{API_BASE_URL}/api/v1/sdk/prompt_run/fetch-by-filter"
            filters_dict = filters.to_dict() if filters is not None else {}
            json = {
                "limit": limit,
                **filters_dict,
            }
            json = {k: v for k, v in json.items() if v is not None}
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json=json,
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            inferences = response.json()["data"]["inferences"]
            return list(map(lambda x: AthinaInference(**x), inferences))
        except Exception as e:
            print("Exception fetching inferences", e)
            pass

    @staticmethod
    def log_usage(eval_name: str, run_type: str):
        """
        Logs a usage event to Posthog via Athina.
        """
        if not AthinaApiKey.is_set():
            return
        try:
            endpoint = f"{API_BASE_URL}/api/v1/sdk/log-usage"
            requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json={
                    "sdkVersion": SDK_VERSION,
                    "evalName": eval_name,
                    "run_type": run_type,
                },
            )
        except Exception as e:
            # Silent failure is ok here.
            pass

    @staticmethod
    @retry(wait_fixed=500, stop_max_attempt_number=3)
    def log_eval_results(
        athina_eval_result_create_many_request: List[AthinaEvalResultCreateRequest],
    ):
        """
        Logs eval results to Athina
        """
        try:
            # Construct eval update requests
            endpoint = f"{API_BASE_URL}/api/v1/eval_result"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json=athina_eval_result_create_many_request,
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()
        except Exception as e:
            print(
                f"An error occurred while posting eval results",
                str(e),
            )
            raise
    
    @staticmethod
    def create_dataset(
        dataset: Dict
    ):
        """
        Creates a dataset by calling the Athina API
        """
        try:
            endpoint = f"{API_BASE_URL}/api/v1/dataset_v2"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json=dataset,
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()['data']['dataset']
        except Exception as e:
            raise
    
    @staticmethod
    def fetch_dataset_rows(
        dataset_id: str,
        number_of_rows: Optional[int] = None
    ):
        """
        Fetch the dataset rows by calling the Athina API

        """
        try:
            if number_of_rows is None:
                number_of_rows = 20
            endpoint = f"{API_BASE_URL}/api/v1/dataset_v2/fetch-by-id/{dataset_id}?offset=0&limit={number_of_rows}&include_dataset_rows=true"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers()
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()['data']['dataset_rows']
        except Exception as e:
            raise

    @staticmethod
    def add_dataset_rows(dataset_id: str, rows: List[Dict]):
        """
        Adds rows to a dataset by calling the Athina API.

        Parameters:
        - dataset_id (str): The ID of the dataset to which rows are added.
        - rows (List[Dict]): A list of rows to add to the dataset, where each row is represented as a dictionary.

        Returns:
        The API response data for the dataset after adding the rows.

        Raises:
        - CustomException: If the API call fails or returns an error.
        """
        try:
            endpoint = f"{API_BASE_URL}/api/v1/dataset_v2/{dataset_id}/add-rows"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json={"dataset_rows": rows},
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get('details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()['data']
        except Exception as e:
            raise

    @staticmethod
    def create_eval_request(
        athina_eval_request_create_request: AthinaEvalRequestCreateRequest,
    ):
        """
        Create eval request
        """
        try:
            endpoint = f"{API_BASE_URL}/api/v1/eval_request"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json=athina_eval_request_create_request,
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()
        except Exception as e:
            print(
                f"An error occurred while creating eval request",
                str(e),
            )
            raise

    def log_eval_performance_report(
        self, eval_request_id: str, report: EvalPerformanceReport
    ):
        """
        Logs the performance metrics for the evaluator.
        """
        try:
            endpoint = f"{API_BASE_URL}/api/v1/eval_performance_report"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json={
                    "eval_request_id": eval_request_id,
                    "true_positives": report["true_positives"],
                    "false_positives": report["false_positives"],
                    "true_negatives": report["true_negatives"],
                    "false_negatives": report["false_negatives"],
                    "accuracy": report["accuracy"],
                    "precision": report["precision"],
                    "recall": report["recall"],
                    "f1_score": report["f1_score"],
                    "runtime": report["runtime"],
                    "dataset_size": report["dataset_size"],
                },
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()
        except Exception as e:
            print(
                f"An error occurred while posting eval performance report",
                str(e),
            )
            raise

    @staticmethod
    def log_experiment(
        eval_request_id: str,
        experiment: AthinaExperiment,
    ):
        """
        Logs the experiment metadata to Athina.
        """
        try:
            endpoint = f"{API_BASE_URL}/api/v1/experiment"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json={
                    "eval_request_id": eval_request_id,
                    "experiment_name": experiment["experiment_name"],
                    "experiment_description": experiment["experiment_description"],
                    "language_model_provider": experiment["language_model_provider"],
                    "language_model_id": experiment["language_model_id"],
                    "prompt_template": experiment["prompt_template"],
                    "dataset_name": experiment["dataset_name"],
                },
            )
            print(response.status_code)
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()
        except Exception as e:
            print(
                f"An error occurred while posting experiment metadata",
                str(e),
            )
            raise
    
    @staticmethod
    def log_eval_results_with_config(eval_results_with_config: dict):
        try:
            endpoint = f"{API_BASE_URL}/api/v1/eval_run/log-eval-results-sdk"
            response = requests.post(
                endpoint,
                headers=AthinaApiService._headers(),
                json=eval_results_with_config,
            )
            if response.status_code == 401:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = 'please check your athina api key and try again'
                raise CustomException(error_message, details_message)
            elif response.status_code != 200 and response.status_code != 201:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(error_message, details_message)
            return response.json()
        except Exception as e:
            raise



import time
from typing import List
from ..evals import BaseEvaluator
from .exception import AthinaGuardException
from concurrent.futures import ThreadPoolExecutor, as_completed


def guard(suite: List[BaseEvaluator], **kwargs):
    # Define the maximum number of threads to use
    max_workers = 10  # Adjust based on your needs and environment
    start_time = time.perf_counter()
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        # Submit all evaluation functions to the executor
        future_to_eval = {executor.submit(eval.guard, **kwargs): eval for eval in suite}

        for future in as_completed(future_to_eval):
            eval = future_to_eval[future]
            try:
                guard_result = future.result()
                passed = guard_result.passed
                reason = guard_result.reason
                runtime = guard_result.runtime
                if passed:
                    print(f"{eval.display_name}: Passed in {runtime}ms - {reason}")
                else:
                    print(f"{eval.display_name}: Failed in {runtime}ms - {reason}")
                    raise AthinaGuardException(f"{eval.display_name} failed: {reason}")
            except Exception as exc:
                raise exc

    end_time = time.perf_counter()
    response_time_ms = (end_time - start_time) * 1000
    print(f"Guard completed in {response_time_ms}ms")



from typing import List, Optional

from athina.llms.abstract_llm_service import AbstractLlmService
from ..llm_evaluator import LlmEvaluator
from athina.metrics.metric_type import MetricType
from athina.evals.eval_type import LlmEvalTypeId


class GradingCriteria(LlmEvaluator):
    """
    This evaluator checks if the response is correct according to a provided `grading_criteria`.
    """

    USER_MESSAGE_TEMPLATE = """
    ### GRADING CRITERIA ###
    {grading_criteria}

    ### EXAMPLES ###
    {examples}

    ### RESPONSE TO EVALUATE ###
    {response}
    """
    _examples = []
    grading_criteria = None
    def __init__(self, 
        grading_criteria: str, 
        model: Optional[str] = None,
        llm_service: Optional[AbstractLlmService] = None
    ):
        if grading_criteria is None:
            raise Exception("Eval is incorrectly configured: grading_criteria is required for GradingCriteria evaluator")
        self.grading_criteria = grading_criteria
        super().__init__(
            model=model,
            llm_service=llm_service
        )

    @property
    def name(self):
        return LlmEvalTypeId.GRADING_CRITERIA.value

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.PASSED.value]

    @property
    def display_name(self):
        return "Response matches Grading Criteria"

    @property
    def default_model(self):
        return "gpt-4-1106-preview"

    @property
    def required_args(self):
        return ["response"]

    @property
    def examples(self):
        return self._examples
    
    def to_config(self) -> Optional[dict]:
        return {
            "grading_criteria": self.grading_criteria
        }
    
    def is_failure(self, result) -> Optional[bool]:
        return bool(str(result).lower() == "fail")

    def _user_message(self, response, **kwargs) -> str:
        """
        Generates data for evaluation.

        :param response: llm response
        :return: A dictionary with formatted data for evaluation
        """
        return self.USER_MESSAGE_TEMPLATE.format(
            examples=self._examples_str(),
            grading_criteria=self.grading_criteria,
            response=response
        )


from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Optional, Dict
from athina.helpers.logger import logger
from athina.helpers.athina_logging_helper import AthinaLoggingHelper
from athina.helpers.dataset_helper import generate_unique_dataset_name, generate_eval_display_name
from athina.interfaces.data import DataPoint
from athina.interfaces.result import BatchRunResult, EvalResult, GuardResult
from athina.services.athina_api_service import AthinaApiService
from athina.datasets import Dataset
import traceback

class BaseEvaluator(ABC):
    
    # Abstract properties
    @property
    @abstractmethod
    def name(self) -> str:
        """A unique name identifier for the evaluator."""
        pass

    @property
    @abstractmethod
    def display_name(self) -> str:
        """A display name for the evaluator."""
        pass

    @property
    @abstractmethod
    def metric_ids(self) -> List[str]:
        """The metric computed by the evaluator."""
        pass

    @property
    @abstractmethod
    def required_args(self) -> List[str]:
        """A list of required arguments for the evaluator."""
        pass

    @property
    @abstractmethod
    def examples(self):
        """A list of examples for the evaluator."""
        pass

    @abstractmethod
    def is_failure(self, *args) -> Optional[bool]:
        """A method to determine if the evaluation failed."""
        pass

    @abstractmethod
    def _evaluate(self, **kwargs) -> EvalResult:
        """The method that performs the evaluation."""
        pass
    
    def to_config(self) -> Optional[Dict]:
        return None

    # Common methods
    def _examples_str(self) -> str:
        return "" if self.examples is None else "\n".join(map(str, self.examples))


    def validate_args(self, **kwargs) -> None:
        """
        Validates that all required arguments are present and not None.
        """
        for arg in self.required_args:
            if arg not in kwargs:
                raise ValueError(f"Missing required argument: {arg}")
            elif kwargs[arg] is None:
                raise ValueError(f"{arg} cannot be None")

    def _validate_batch_args(self, data: List[DataPoint]) -> bool:
        """
        Validates that each entry in the batch has all the required arguments,
        and none of the arguments is None.
        """
        for i, entry in enumerate(data):
            for arg in self.required_args:
                if arg not in entry:
                    raise ValueError(
                        f"Data at index {i} is missing required argument: {arg}"
                    )
                elif entry[arg] is None:
                    raise ValueError(
                        f"Data at index {i} has required argument {arg} set to None"
                    )
        return True

    def _log_evaluation_request(self, data) -> Optional[str]:
        """
        Logs usage to Athina for analytics and creates an evaluation request.
        """
        eval_request = None
        try:
            eval_request = AthinaLoggingHelper.create_eval_request(
                eval_name=self.name, request_data={"data": data}, request_type="batch"
            )
        except Exception as e:
            pass
        return eval_request


    def _log_evaluation_results(
        self, eval_request_id: Optional[str], eval_results: List[EvalResult], org_id: Optional[str] = None, workspace_slug: Optional[str] = None
    ):
        """
        Logs the evaluation results to Athina if the eval_request_id is available.
        """
        if eval_request_id:
            try:
                AthinaLoggingHelper.log_eval_results(
                    eval_request_id=eval_request_id,
                    eval_results=eval_results,
                    org_id=org_id,
                    workspace_slug=workspace_slug
                )
            except Exception as e:
                pass

    def run(self, **kwargs) -> BatchRunResult:
        """
        Run the LLM evaluator, and log results to Athina.
        """
        AthinaApiService.log_usage(eval_name=self.name, run_type="batch")
        eval_request = self._log_evaluation_request(kwargs)
        eval_result = self._evaluate(**kwargs)
        self._log_evaluation_results(
            eval_request_id=eval_request["eval_request"]["id"], eval_results=[eval_result], org_id=eval_request["eval_request"]["org_id"], workspace_slug=eval_request["eval_request"]["workspace_slug"]
        )

        return BatchRunResult(
            eval_request_id=eval_request["eval_request"]["id"],
            eval_results=[eval_result],
        )

    def guard(self, **kwargs):
        """
        Guard
        """
        eval_result = self._evaluate(**kwargs)
        passed = not eval_result["failure"]
        reason = eval_result["reason"]
        runtime = eval_result["runtime"]
        return GuardResult(passed=passed, reason=reason, runtime=runtime)

    def _run_batch_generator_async(
        self, data: List[DataPoint], max_parallel_evals: int
    ):
        with ThreadPoolExecutor(max_workers=max_parallel_evals) as executor:
            # Submit all tasks to the executor and store them with their original index
            future_to_index = {
                executor.submit(self._evaluate, **entry): i
                for i, entry in enumerate(data)
            }

            # Create a list to store results in the original order
            results = [None] * len(data)

            for future in as_completed(future_to_index):
                index = future_to_index[future]
                try:
                    results[index] = future.result()
                except Exception as e:
                    entry = data[index]
                    logger.error(f"Error running batch async {entry}: {e}")
                    traceback.print_exc()
                    results[index] = None

            return results

    def _run_batch_generator(self, data: List[DataPoint]):
        """
        Generator function for running a batch of evaluations.
        Iterates over a dataset, and runs the evaluator on each entry.
        """
        for entry in data:
            try:
                yield self._evaluate(**entry)
            except Exception as e:
                logger.error(f"Error evaluating entry {entry}: {e}")
                traceback.print_exc()
                yield None

    def _log_dataset_to_athina(self, data: List[DataPoint]) -> Optional[str]:
        """
        Logs the dataset to Athina
        """
        try: 
            dataset = Dataset.create(
                name=generate_unique_dataset_name(),
                rows=data
            )
            return dataset
        except Exception as e:
            print(f"Error logging dataset to Athina: {e}")
            return None
    
    def _log_eval_results_to_athina(self, eval_results: List[EvalResult], dataset_id: str):
        """
        Logs the batch results to Athina
        """
        try:
            eval_config = self.to_config()
            llm_engine = getattr(self, "_model", None)
            AthinaLoggingHelper.log_eval_results_with_config(
                eval_results_with_config={
                    "eval_results": eval_results,
                    "development_eval_config": {
                        "eval_type_id": self.name,
                        "eval_display_name": generate_eval_display_name(self.display_name),
                        "eval_config": eval_config,
                        "llm_engine": llm_engine
                    }
                },
                dataset_id=dataset_id
            )
        except Exception as e:
            print(f"Error logging eval results to Athina: {e}")
            pass
    
    def run_batch(
        self, data: List[DataPoint], max_parallel_evals: int = 5
    ) -> BatchRunResult:
        """
        Runs the evaluator on a batch of data.
        """
        # Log usage to Athina for analytics
        AthinaApiService.log_usage(eval_name=self.name, run_type="batch")

        # Run the evaluations
        if max_parallel_evals > 1:
            eval_results = self._run_batch_generator_async(data, max_parallel_evals)
        else:
            eval_results = list(self._run_batch_generator(data))

        # Create the Dataset
        dataset = self._log_dataset_to_athina(data)
        if dataset:
            self._log_eval_results_to_athina(eval_results, dataset.id)
            print(f"You can view your dataset at: {Dataset.dataset_link(dataset.id)}")
        

        return BatchRunResult(
            eval_results=eval_results,
        )



from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Callable, Iterable
from athina.steps.base import Step


class Map(Step):
    """
    Step that applies a function to each item in the input data.

    Attributes:
        fn (Callable[[Any], Any]): Function to apply to each item.
    """

    fn: Callable[[Any], Any]

    def execute(self, input_data: Any) -> List[Any]:
        """Apply a function to each item in the input data."""
        if not isinstance(input_data, Iterable):
            raise ValueError("Input data must be an iterable")
        results = list(map(self.fn, input_data))
        return results



# Step to chat with OpenAI's Assistant API.
from typing import Union, Dict, Any
from athina.steps import Step
from openai import OpenAI
import os


class OpenAiAssistant(Step):
    """
    Step that chats with OpenAI's Assistant API.

    Attributes:
        assistant_id: The assistant ID to be used.
        openai_api_key: OpenAI's API Key.
        input_column: The row's column to classify.
    """
    assistant_id: str
    openai_api_key: str
    input_column: str
    client: any = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, assistant_id: str, openai_api_key: str, input_column: str):
        super().__init__(assistant_id=assistant_id, openai_api_key=openai_api_key, input_column=input_column)
        self.client = OpenAI(api_key=openai_api_key)

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Calls OpenAI's Assistant API and returns the response."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        input_text = input_data.get(self.input_column, None)
        
        if input_text is None:
            return None
        try:
            # Create a thread
            thread = self.client.beta.threads.create()

            # Add input_text to the thread
            self.client.beta.threads.messages.create(
                thread_id=thread.id,
                role='user',
                content=input_text
            )

            # Run the assistant
            run = self.client.beta.threads.runs.create(
                thread_id=thread.id,
                assistant_id=self.assistant_id
            )

            # Wait for the run to complete
            while run.status not in ["completed", "failed"]:
                run = self.client.beta.threads.runs.retrieve(thread_id=thread.id, run_id=run.id)

            # Handle failed case
            if run.status == "failed":
                return {
                    "status": "error",
                    "data": "The assistant run failed."
                }
            
            # Retrieve the assistant's response
            messages = self.client.beta.threads.messages.list(thread_id=thread.id)
            
            # Check and return the assistant's response based on format
            for message in messages.data:
                if message.role == "assistant":
                    for content in message.content:
                        if content.type == "text":
                            return {"status": "success", "data": content.text.value}
                        elif content.type == "json":
                            return {"status": "success", "data": content.json.value}
            return {
                "status": "success",
                "data": None,
            }
        except Exception as e:
            return {
                "status": "error",
                "data": str(e),
            }
